/*
  This flex scanner is still C-fashion.
*/
%{                                           /* -*- C++ -*- */
# include <cstdlib>
# include <errno.h>
# include <limits.h>
# include <string>
# include "calc++-driver.h"
# include "calc++-parser.tab.hpp"

  /* Work around an incompatibility in flex (at least versions
     2.5.31 through 2.5.33): it generates code that does
     not conform to C89.  See Debian bug 333231
     <http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=333231>.  */
# undef yywrap
# define yywrap() 1

  /* By default yylex returns int, we use token_type.
     Unfortunately yyterminate by default returns 0, which is
     not of token_type.  */
#define yyterminate() return token::END
%}

%option noyywrap nounput batch debug

int [0-9]+
var [a-z][a-z,0-9]*
blank [ \t]
float [0-9]+.[0-9]+
string '[A-Za-z_\- 0-9]*'
relation [A-Z]+[A-Za-z0-9]*
aop [+-*\%]
bop [&|]
bool true|false|TRUE|FALSE


%%
{blank}+ yylloc->step();
[\n]+ yylloc->lines(yyleng); yylloc->step();

%{
  typedef datalog::DatalogParser::token token;
%}


{int} {
          long n = strtol (yytext, NULL, 10);
          yylval->ival = n; 
          return token::INUM;
      }

{var} {
        string *s = new string(yytext);
      yylval->sval = s;
      return token::VAR
}
      
{float} {
        double t = strtod(yytext, NULL);
        yylval->fval = t;
        return token::FNUM;
        }
{string} {
          // string contains a ' '
        string *s = new string(yytext);
        yylval->sval = s;
        return token::STRING;
         }
{relation} {
       string *s = new string(yytext);
       yylval->sval = s;
       return token::RELSYM;
}
{aop} { 
  string *s = new string(yytext);
  yylval->sval = s;
  return token::AOP;
}
{bop} { 
  string *s = new string(yytext);
  yylval->sval = s;
  return token::BOP;
}
{bool} {
  if(yytext[0] == 'T' || yytext[0] =='t') {
    yylval->bval = true;
  } else
  {
    yylval->bval = false;
  }
  return token::BOOL;
  
}

%%

void DatalogParser::scan_begin ()
{
  yy_flex_debug = trace_scanning;
  if (filename == "-")
    yyin = stdin;
  else if (!(yyin = fopen (filename.c_str (), "r")))
  {
    cerr<<std::string ("cannot open ")<< filename <<endl;
    exit (1);
  }

}

void DatalogParser::scan_end ()
{
  fclose (yyin);
}