#include "datalogtree.h"
#include "treedlgtool.h"
#include "dlogrel.h"
#include <queue>

namespace datalog {

	TreeNodeTypeMap DlogTreeNode::nodetypemap;
	bool DlogTreeNode::inited = false;
	void DlogTreeNode::InitNodeTypeMap() {
	    if (!inited) {
	        DlogTreeNode::nodetypemap[N_ROOT] = string("N_ROOT");
	        DlogTreeNode::nodetypemap[N_RULE] = string("N_RULE");
	        DlogTreeNode::nodetypemap[N_HEAD] = string("N_HEAD");
	        DlogTreeNode::nodetypemap[N_BODY] = string("N_BODY");
	        DlogTreeNode::nodetypemap[N_VAR] = string("N_VAR");
	        DlogTreeNode::nodetypemap[N_NUMCONST] = string("N_NUMCONST");
	        DlogTreeNode::nodetypemap[N_STRCONST] = string("N_STRCONST");
	        DlogTreeNode::nodetypemap[N_PRED] = string("N_PRED");
	        DlogTreeNode::nodetypemap[N_RELATION] = string("N_RELATION");
	        DlogTreeNode::nodetypemap[N_AGG] = string("N_AGG");
	        DlogTreeNode::nodetypemap[N_RELSYM] = string("N_RELSYM");
	        DlogTreeNode::nodetypemap[N_AGGFUNC] = string("N_AGGFUNC");
	        DlogTreeNode::nodetypemap[N_EXPR] = string("N_EXPR");
	        DlogTreeNode::nodetypemap[N_BOOLOP] = string("N_BOOLOP");
	        DlogTreeNode::nodetypemap[N_MATHOP] = string("N_MATHOP");
	        DlogTreeNode::nodetypemap[N_SUBGOAL] = string("N_SUBGOAL");
	        DlogTreeNode::nodetypemap[N_PREDLIST] = string("N_PREDLIST");
	        //
	        DlogTreeNode::nodetypemap[N_UNDEFINED] = string("N_UNDEF");
	        inited = true;
	    }
	}
	
	DlogTreeNode* DlogTreeNode::Read(istream& in, int& cs) {
	    //Read from in and create a node object
	    DlogNodeValue v;
	    int inttype;
	    in>>inttype;
	    in>>cs;
	    DlogTreeNodeType t = (DlogTreeNodeType) inttype;
	    v.read(in);
	    DlogTreeNode* n = new DlogTreeNode(t, v);
	    return n;
	}
	
	DlogTreeNode::DlogTreeNode() {
	    Init();
	}
	
	DlogTreeNode::DlogTreeNode(const DlogTreeNode& node) : value(node.getValue()) {
	    type = node.getNodeType();
	    /*
	    int n = node.getChildrenSize();
	    for(int i=0 ; i<n; ++i) {
	      children.push_back(node.children[i]);
	    }
	    */
	    parent = node.parent;
	}
	
	DlogTreeNode::DlogTreeNode(const DlogTreeNodeType _type, DlogNodeValue v):
	type(_type), value(v), parent(NULL) {
	    InitNodeTypeMap();
	    children.clear();
	}
	DlogTreeNode::~DlogTreeNode() {
	    clear();
	}
	
	void DlogTreeNode::setType (DlogTreeNodeType t) {
	    type = t;
	}
	void DlogTreeNode::setValue(DlogNodeValue v) {
	    value = v;
	}
	void DlogTreeNode::setValue(const string& v) {
	    // set the value string to v.
	    value.setValue(v);
	}
	const DlogNodeValue& DlogTreeNode::getValue() const {
	    return value;
	}
	bool DlogTreeNode::equal(const DlogTreeNode& rhs) const {
	    //check only the node value, not the subtree or even fanout. leave that to DatalogOps::Identical()
	    if (type != rhs.type) return false;
	    if (!value.equal(rhs.getValue())) return false;
	        return true;
	    }
	    bool DlogTreeNode::equal(const string& strval) const {
	            string s;
	            value.getValue(&s);
	        if (s.compare(strval) == 0) {
	        return true;
	    }
	return false;
	}
	
	const DlogTreeNodeType DlogTreeNode::getNodeType() const {
	    return type;
	}
	
	const DlogTreeNode& DlogTreeNode::getChild(const int idx) const  {
	if (idx >= getChildrenSize() ) {
	cerr<<"Invalid Children index"<<endl;
	return *(new DlogTreeNode());
	}
	
	return *children[idx];
	}
	
	const int DlogTreeNode::getChildrenSize() const {
	    return children.size();
	}
	
	// return the effective children size
	const int DlogTreeNode::getEffChildrenSize() const {
	    int s = getChildrenSize();
	    int t = 0;
	for (int i=0; i<s ; i++) {
	if (children[i] != NULL) ++t;
	}
	return t;
	}
	
	DlogTreeNode* DlogTreeNode::retChild(const int idx) const {
	if (idx >= getChildrenSize() ) {
	cerr <<" Invalid Children index" <<endl;
	return NULL;
	}
	
	return children[idx];
	}
	
	const DlogTreeNode& DlogTreeNode::getParent() const {
	if (parent) {
	return *parent;
	}
	
	cerr<<"parent is null, are you at root?"<<endl;
	return *(new DlogTreeNode());
	
	}
	
	DlogTreeNode* DlogTreeNode::retParent() const {
	    return parent;
	}
	
	void DlogTreeNode::setParent(const DlogTreeNode* parent) {
	    if (parent != NULL) {
	        this->parent = const_cast<DlogTreeNode*> (parent);
	    } else {
	        this->parent = NULL;
	    }
	}
	
	int DlogTreeNode::appendChild(DlogTreeNode* node) {
	    // The tree take the ownership of the appended node.
	    children.push_back(node);
	    node->setParent(this);
	    // return the index of the inserted node.
	    return children.size()-1;
	}
	
	int DlogTreeNode::pushChild(DlogTreeNode* node) {
	    //Push to the head of the children list, can take linear time
	    children.insert(children.begin(), node);
	    node->setParent(this);
	    // always return 0, the index of pushed in node.
	    return 0;
	}
	// remove a child
	int DlogTreeNode::detachChild(const int idx, const int mode) {
	    /* detach a child, 3 modes:
	    mode 0: remove from children and delete
	    mode 1: remove from children but don't delete it
	    mode 2: detach from children, delete and set that entry to NULL
	    mode 3: detach from children, set entry to NULL but don't delete
	      
	    for mode 1 and 3, user should call retChildren to avoid memory leak.
	    */
	    if (idx >= getChildrenSize()) return -1;
	    Children::iterator it = children.begin();
	    for (int i =0 ; i< idx && it!=children.end(); ++i, ++it) {};
	    DlogTreeNode* p = children[idx];
	
	    switch (mode) {
	    case 0:
	        children.erase(it);
	        delete p;
	        break;
	    case 1:
	        children.erase(it);
	        if (p!=NULL) p->setParent(NULL);
	        break;
	    case 2:
	        children[idx] = NULL;
	        delete p;
	        break;
	    case 3:
	        children[idx] = NULL;
	        if (p!=NULL) p->setParent(NULL);
	        break;
	    }
	    return 0;
	}
	/*
	Replace child does not de-allocate the child being replaced.
	the replaced part is deallocated.
	*/
	int DlogTreeNode::replaceSubtree(const int idx, DlogTreeNode* node) {
	    DlogTreeNode* p = children.at(idx);
	    if (p == node) return -1;
	    node->setParent(&(p->getParent()));
	    p->setParent(NULL);
	    children[idx] = node;
	    delete p;
	    return 0;
	}
	
	void DlogTreeNode::PrintTreeNode(ostream& out, int indent) const {
	    for (int i=0; i<indent; i++) out<<" ";
	    out<< "(" << nodetypemap[type] << ",";
	    value.Print(out);
	    out<<"):" << children.size()<<endl;
	}
	
	void DlogTreeNode::PrintSubTree(ostream& out, int indent) const {
	        // pre-order print.
	        PrintTreeNode(out, indent);
	        int n = children.size();
	        int increment  = 2;
	    for (int i = 0; i<n; i++) {
	    if (children[i] != NULL)
	            children[i]->PrintSubTree(out, indent+increment);
	        else {
	            for (int j=0; j<indent; j++) {
	                out<<" ";
	            }
	            out<<"[NULL]"<<endl;
	        }
	    }
	}
	
	void DlogTreeNode::PrintSubTreeMid(string& dlog) const {
	    // binary tree mid travalsal
	    if (getChildrenSize() > 0) retChild(0)->PrintSubTreeMid(dlog);
	    string s;
	    getValue().getValue(&s);
	    dlog.append(s);
	    if (getChildrenSize() > 0) retChild(1)->PrintSubTreeMid(dlog);
	    }
	//Recursively release nodes
	    void DlogTreeNode::clear() {
	        const int n = getChildrenSize();
	        for (int i=0; i<n; i++) {
	            if (children[i] != NULL) {
	                children[i]->clear();
	                delete children[i];
	            }
	        }
	        children.clear();
	    }
	
	DlogTreeNode* DlogTreeNode::findNodeInSubTree(DlogTreeNodeType type) {
	    if (this->type == type) {
	        return this;
	    };
	    const int n = getChildrenSize();
	    DlogTreeNode* t = NULL;
	    for (int i=0; i<n; i++) {
	        t = retChild(i)->findNodeInSubTree(type);
	        if (t!=NULL) return t;
	    }
	    return t;
	}
	
	void DlogTreeNode::findAllNodesV(const DlogTreeNode* s,
	                                 vector<DlogTreeNode*>& res) {
	    if (type == s->getNodeType() && value.equal(s->getValue())) {
	        res.push_back(this);
	    }
	    int size = getChildrenSize();
	    for (int i=0; i<size; i++) {
	        if (children[i] != NULL) {
	            children[i]->findAllNodesV(s, res);
	        }
	    }
	}
	
	void DlogTreeNode::findAllNodesT(const DlogTreeNodeType t,
	                                 vector<DlogTreeNode*>& res) {
	    if (type == t) {
	        res.push_back(this);
	    }
	    int size = getChildrenSize();
	    for (int i=0; i<size; i++) {
	        if (children[i] != NULL) {
	            children[i]->findAllNodesT(t, res);
	        }
	    }
	}
	DlogTreeNode* DlogTreeNode::deepCopy() const {
	    //DeepCopy the subtree
	    DlogTreeNode *t = new DlogTreeNode(*this);
	    t->setParent(NULL);
	    int s = getChildrenSize();
	for (int i = 0; i<s; i++) {
	t->appendChild(getChild(i).deepCopy());
	}
	return t;
	}
	
	int DlogTreeNode::findChildbyPointer(DlogTreeNode* ptr) {
	    int size = getChildrenSize();
	    for (int i=0; i<size; ++i) {
	        if (children[i] == ptr) {
	            return i;
	        }
	    }
	    return -1;
	}
	int DlogTreeNode::findSelfIdx() {
	    return parent->findChildbyPointer(this);
	}
	
	void DlogTreeNode::Dump(ostream& out) const {
	    out<<type<<" "<<children.size();
	    value.dump(out);
	}
	
	void DlogTreeNode::pack() {
	    vector<DlogTreeNode*> t;
	    int s = children.size();
	    for ( int i=0; i<s; ++i) {
	        if (children[i] != NULL) t.push_back(children[i]);
	    }
	    children.assign(t.begin(), t.end());
	}
	
	// For datalog Tree
	
	DatalogTree::DatalogTree()
	: LangTree(), root(N_ROOT) {
	
	}
	
	DatalogTree::DatalogTree(const DatalogTree& t) {
	    // deep copy a tree.
	    Init();
	    int s = t.getRoot()->getChildrenSize();
	    for ( int i=0; i<s; ++i) {
	        root.appendChild(t.getRoot()->getChild(i).deepCopy());
	    }
	}
	
	DatalogTree::~DatalogTree() {
	    // Recusively delete all the children
	    clearTree();
	}
	
	void DatalogTree::Init() {
	    root = DlogTreeNode(N_ROOT);
	
	}
	
	DlogTreeNode* DatalogTree::retRoot() {
	    return &root;
	}
	
	void DatalogTree::clearTree() {
	    // Recursively clear the tree, release all nodes;
	    root.clear();
	    root.setType(N_UNDEFINED);
	}
	
	DlogTreeNode* DatalogTree::findNode(DlogTreeNodeType type, DlogTreeNode* start) const {
	    //Depth first search, start is not included in the search.
	    if (start == NULL) start = const_cast<DlogTreeNode*>(&root);
	    DlogTreeNode* t = NULL;
	    int n = start->getChildrenSize();
	    for (int i=0; i<n; i++) {
	        if (start->retChild(i) != NULL) {
	            t = start->retChild(i)->findNodeInSubTree(type);
	            if (t!=NULL) return t;
	        }
	    }
	    return t;
	}
	
	// Print datalog from tree
	void DatalogTree::printDatalog() const {
	    TreeDlgTool c;
	    string dlog;
	    c.convert(this, dlog);
	    cout<<dlog<<endl;
	}
	
	void DatalogTree::printDatalog(ostream& out) const {
	    TreeDlgTool c;
	    string dlog;
	    c.convert(this, dlog);
	    out<<dlog<<endl;
	}
	
	void DatalogTree::printTree() const {
	    root.PrintSubTree(cout, 0);
	}
	
	//Substitude variables, used for mapping var to var and var to constant
	int DatalogTree::substitude(const DlogTreeNodeType type, const string& value, DlogTreeNode* target, DlogTreeNode* start) {
	    if (start == NULL) start = &root;
	    int cnt = 0;
	    int n = start->getChildrenSize();
	    for (int i=0; i<n; i++) {
	        if (start->retChild(i) != NULL) {
	            cnt+=substitude(type, value, target, start->retChild(i));
	        }
	    }
	    if (start->getNodeType() == type) {
	        //Compare the value
	        string s;
	        start->getValue().getValue(&s);
	        if (s.compare(value) == 0) {
	            //Replace.
	            start->setType(target->getNodeType());
	            start->setValue(target->getValue());
	            ++cnt;
	        }
	    }
	    return cnt;
	}
	
	int DatalogTree::findAllNodesV(const DlogTreeNode *s,
	                               vector<DlogTreeNode*>& res,
	                               DlogTreeNode* start) {
	    res.clear();
	    if (start == NULL) start = retRoot();
	    start->findAllNodesV(s, res);
	    return res.size();
	}
	
	int DatalogTree::findAllNodesT(const DlogTreeNodeType t,
	                               vector<DlogTreeNode*>& res,
	                               DlogTreeNode* start) {
	    res.clear();
	    if (start == NULL) start = retRoot();
	    start->findAllNodesT(t, res);
	    return res.size();
	}
  int DatalogTree::findRelbyName(const string& name,
                                 vector<int>& res,
                                 DlogTreeNode* start) const {
     DlogTreeNode* subg;
     if(start == NULL) {
       subg = getSubG();
     } else if(start->getNodeType() == N_SUBGOAL) {
         subg = start;
     } else {
       cerr<<"start search from a N_SUBGOAL node"<<endl;
       return -1;
     }
     int s = subg->getChildrenSize();
     for(int i=0; i<s; ++i) {
       if(subg->retChild(i)->getNodeType() == N_RELATION) {
         DlogRel r(subg->retChild(i));
         if(r.getRSymbol().compare(name) == 0) {
           res.push_back(i);
         }
       }
     }
     return res.size();
  }
	
	DlogTreeNode* DatalogTree::getHead() const {
	    return root.retChild(0);
	}
	
	DlogTreeNode* DatalogTree::getBody() const {
	    return root.retChild(1);
	}
	
	void DatalogTree::Dump(ostream& out) const {
	    //breadth first dumping;
	    const DlogTreeNode* r = getRoot();
	    queue<const DlogTreeNode*> q;
	    q.push(r);
	while (!q.empty()) {
	r = q.front();
	    q.pop();
	    r->Dump(out);
	    for (int i=0; i<r->getChildrenSize(); ++i) {
	        q.push(r->retChild(i));
	    }
	}
	}
	
	DatalogTree* DatalogTree::Read(istream& in) {
	    DatalogTree *t = new DatalogTree();
	    int cs;
	    int nc; // number of child to attach
	    queue<int> qs;
	    queue<DlogTreeNode*> qn;
	    DlogTreeNode *root = DlogTreeNode::Read(in,cs);
	    qs.push(cs);
	    qn.push(t->retRoot());
	    delete root;
	    DlogTreeNode* pn;
	    DlogTreeNode* tmpn;
	    while (!qs.empty()) {
	        nc = qs.front();
	        qs.pop();
	        pn = qn.front();
	        qn.pop();
	        for (int i=0; i<nc; ++i) {
	            tmpn = DlogTreeNode::Read(in, cs);
	            qs.push(cs);
	            qn.push(tmpn);
	            pn->appendChild(tmpn);
	        }
	    }
	    return t;
	}
	
	DlogTreeNode* DatalogTree::getSubG() const {
		return root.retChild(1)->retChild(0);
	}

	DatalogTree* DatalogTree::DeepCopy() const {
		DatalogTree *t = new DatalogTree();
		int ts = root.getChildrenSize();
		for(int i=0; i<ts; ++i) {
			t->retRoot()->appendChild(root.retChild(i)->deepCopy());
		}
		return t;	
	}
} /*namespace */
