
#include "dlognodevalue.h"

namespace datalog {

    DlogNodeValue::DlogNodeValue() {
        type = UNDEF;
        value.sval=NULL;
    }

    DlogNodeValue::DlogNodeValue(const DlogNodeValue& v):type(UNDEF) {
        setValue(v);
    }

    DlogNodeValue::~DlogNodeValue() {
        if (type == STRING) {
            delete value.sval;
        }
    }

    DlogNodeValue& DlogNodeValue::operator =(const DlogNodeValue& rhs) {
        if (&rhs == this) return *this;
        setValue(rhs);
        return *this;
    }

    void DlogNodeValue::getValue(string* v) const {
        // Handle value string
    if (type == STRING) {
    v->assign(*value.sval);
        return;
    }
    if (type == INT) {
    char s[20];
        sprintf(s, "%d", value.ival);
        v->assign(s);
        return;
    }
    if (type == DOUBLE) {
    char s[20];
        sprintf(s, "%lf", value.fval);
        v->assign(s);
        return;
    }
    if (type == BOOLEAN) {
    if (value.bval) {
            v->assign("true");
        } else {
            v->assign("false");
        }
        return;
    }
    }

    void DlogNodeValue::setValue(const string& s) {
        if (type == STRING) {
            if (value.sval) {
                value.sval->assign(s);
            } else {
                value.sval = new string();
                value.sval->assign(s);
            }
        } else {
            type = STRING;
            value.sval = new string();
            value.sval->assign(s);
        }
    }

    void DlogNodeValue::setValue(const char* s) {
        setValue(string(s));
    }

    void DlogNodeValue::setValue(const DlogNodeValue& v) {
        switch (v.getValueType()) {
        case INT: {
            int i;
            v.getValue(&i);
            setValue(i);
            break;
        }
        case DOUBLE: {
            double f;
            v.getValue(&f);
            setValue(f);
            break;
        }
        case STRING: {
            string s;
            v.getValue(&s);
            setValue(s);
            break;
        }
        case BOOLEAN: {
            bool b;
            v.getValue(&b);
            setValue(b);
            break;
        }
        case UNDEF:
            break;
        }
    }
    void DlogNodeValue::Print(ostream& out) const {
      switch (type) {
        case INT :
           out<<value.ival;
        break;
        case DOUBLE :
          out<<value.fval;
        break;
        case STRING :
            out<<*value.sval;
        break;
        case BOOLEAN :
            if (value.bval) out<<"TRUE";
            else out<<"FALSE";
        break;
        case UNDEF :
            out<<"EMPTY VALUE";
      };
    }

    bool DlogNodeValue::equal(const DlogNodeValue& rhs) const {
        if (type != rhs.type) return false;
        switch (type) {
          case INT :
              return value.ival == rhs.getValues().ival;
          case DOUBLE :
              return value.fval == rhs.getValues().fval;
          case STRING :
              return value.sval->compare(*rhs.getValues().sval) == 0;
          case BOOLEAN :
              return value.bval ==rhs.getValues().bval;
          case UNDEF :
              return true;
        }
    }

    void DlogNodeValue::dump(ostream& out) const {
        out<<" "<<type<<" ";
        switch (type) {
          case INT :
              out<<value.ival;
              break;
          case DOUBLE :
              out<<value.fval;
              break;
          case STRING :
              out<<value.sval->length()<<" "<<*value.sval;
              break;
          case BOOLEAN :
              out<<value.bval;
              break;
          case UNDEF :
              out<<"UNDEF";
              break;
        }
        out<<endl;
    }

    void DlogNodeValue::read(istream& fin) {
        char buf[255];
        int strlen;
        int inttype;
        char c;
        fin>>inttype;
        type = (ValueTypes)inttype;
        switch (type) {
        case INT :
            fin>>value.ival;
            fin.getline(buf, 255);
            break;
        case DOUBLE :
            fin>>value.fval;
            fin.getline(buf, 255);
            break;
        case STRING :
            fin>>strlen;
            fin.getline(buf, 255);
            buf[strlen+1]='\0';
            this->setValue(buf+1);
            break;
        case BOOLEAN :
            fin>>value.bval;
            fin.getline(buf, 255);
            break;
        case UNDEF :
            fin.getline(buf, 255);
        }
    }
    
    
    void DlogNodeValue::gstr() {
      if(type == STRING && value.sval != NULL) {
        delete value.sval;
        value.sval = NULL;
      }
    };

}
