
#ifndef DATALOGDLOGNODEVALUE_H
#define DATALOGDLOGNODEVALUE_H

#include <string>
#include <iostream>

using namespace std;
namespace datalog {

/**
Base class for DlogNodeValue

	@author xujian <shelbyxu@gmail.com>
*/
typedef union values
{
  int ival;
  double fval;
  bool bval;
  string* sval;  
} Values;

typedef enum valueTypes {
  INT = 0,
  DOUBLE = 1,
  STRING = 2,
  BOOLEAN = 3,
  UNDEF = 100
} ValueTypes;

class DlogNodeValue{
public:
    DlogNodeValue();
    DlogNodeValue(const int v) :type(UNDEF) {setValue((int) v);}
    DlogNodeValue(const double v):type(UNDEF) {setValue((double) v);}
    DlogNodeValue(const string& s):type(UNDEF) {setValue(s);}
    DlogNodeValue(const char* s):type(UNDEF) {setValue(s);}
    DlogNodeValue(const bool b):type(UNDEF) {setValue(b);};
    DlogNodeValue(const DlogNodeValue& v);
    virtual ~DlogNodeValue();
    const ValueTypes getValueType() const { return type; } 
    const Values getValues() const {return value; }
    void getValue(int* v) const { *v = value.ival;};
    void getValue(double* v) const {*v = value.fval;};
    void getValue(string* v) const;
    void getValue(bool *v) const {*v = value.bval;};
    void setValue(const int v) { gstr(); value.ival = v; type=INT;};
    void setValue(const double v) {gstr(); value.fval = v; type=DOUBLE;};
    void setValue(const bool v) { gstr(); value.bval = v; type = BOOLEAN;};
    void setValue(const string& s);
    void setValue(const char* s);
    void setValue(const DlogNodeValue& v);
    DlogNodeValue& operator = (const DlogNodeValue& rhs);
    void Print(ostream& out) const;
    bool equal(const DlogNodeValue& rhs) const;
    void dump(ostream& out) const;
    void read(istream& fin);
private:
    ValueTypes type;
    Values value;
    // guard the string allocation. called when the value type is changing.
    void gstr();
};

// Derived class for different value types.

}

#endif
